advent_of_code::solution!(13);

type Board = Vec<Vec<char>>;
type Pair = (usize, usize);

struct Map {
    rows: Board,
    cols: Board,
    width: usize,
    height: usize,
}

impl Map {
    fn find_mirror(&self, board: &Board) -> Option<Pair> {
        let mut matches: Vec<Pair> = vec![];
        let mut start_index = 0;
        while let Some(m) = self.find_row(board, start_index, 0) {
            start_index = m.1;
            matches.push(m);
        }

        let mut res: Vec<Option<Pair>> = vec![];
        for &(r1, r2) in matches.iter() {
            let mut rows_matched = 0;
            let mut found = true;
            loop {
                let start = r1 - rows_matched;
                let end = r2 + rows_matched;
                if board[start] == board[end] {
                    rows_matched += 1;
                } else {
                    found = false;
                    break;
                }
                if start == 0 || end == board.len() - 1 {
                    break;
                }
            }

            if found {
                res.push(Some((r1, r2)));
            }
        }

        if res.len() > 1 {
            panic!("got too many matches {:?}", res);
        } else if res.is_empty() {
            None
        } else {
            res[0]
        }
    }

    fn find_row(&self, board: &Board, index: usize, max_diffs: usize) -> Option<Pair> {
        let row_len = board[0].len();
        for t in index..(board.len() - 1) {
            let mut diff_chars = 0;
            for k in 0..row_len {
                if board[t][k] == board[t + 1][k] {
                    continue;
                }
                diff_chars += 1;
                if diff_chars > max_diffs {
                    break;
                }
            }
            if diff_chars <= max_diffs {
                return Some((t, t + 1));
            }
        }
        None
    }

    fn get_horiz_score(&self) -> Option<usize> {
        if let Some((_r1, r2)) = self.find_mirror(&self.rows) {
            return Some(100 * r2);
        }
        None
    }

    fn get_vert_score(&self) -> Option<usize> {
        if let Some((_c1, c2)) = self.find_mirror(&self.cols) {
            return Some(c2);
        }
        None
    }

    fn part1(&self) -> u32 {
        if self.width < self.height {
            if let Some(n) = self.get_horiz_score() {
                return n as u32;
            }
            if let Some(n) = self.get_vert_score() {
                return n as u32;
            }
        } else {
            if let Some(n) = self.get_vert_score() {
                return n as u32;
            }
            if let Some(n) = self.get_horiz_score() {
                return n as u32;
            }
        }
        0
    }

    fn find_smudge(&self, board: &Board) -> Option<Pair> {
        let mut matches: Vec<Pair> = vec![];
        let mut start_index = 0;
        while let Some(m) = self.find_row(board, start_index, 1) {
            start_index = m.1;
            matches.push(m);
        }

        let mut res: Vec<Option<Pair>> = vec![];
        for &(r1, r2) in matches.iter() {
            let mut rows_matched = 0;
            let mut diff_chars = 0;
            loop {
                let start = r1 - rows_matched;
                let end = r2 + rows_matched;
                if board[start] == board[end] {
                    rows_matched += 1;
                } else {
                    for index in 0..board[start].len() {
                        if board[start][index] != board[end][index] {
                            diff_chars += 1;
                        }
                    }
                    if diff_chars > 1 {
                        break;
                    }
                    rows_matched += 1;
                }
                if start == 0 || end == board.len() - 1 {
                    break;
                }
            }

            if diff_chars == 1 {
                res.push(Some((r1, r2)));
            }
        }

        if res.len() > 1 {
            panic!("got too many matches {:?}", res);
        } else if res.is_empty() {
            None
        } else {
            res[0]
        }
    }

    fn get_horiz_smudge(&self) -> Option<usize> {
        if let Some((_r1, r2)) = self.find_smudge(&self.rows) {
            return Some(100 * r2);
        }
        None
    }

    fn get_vert_smudge(&self) -> Option<usize> {
        if let Some((_c1, c2)) = self.find_smudge(&self.cols) {
            return Some(c2);
        }
        None
    }

    fn part2(&self) -> u32 {
        let hscore = self.get_horiz_smudge();
        let vscore = self.get_vert_smudge();
        if hscore.is_some() && vscore.is_some() {
            panic!("got both scores!");
        }
        if hscore.is_some() {
            hscore.unwrap() as u32
        } else if vscore.is_some() {
            vscore.unwrap() as u32
        } else {
            0
        }
    }
}

fn gen_map(cur: &Board) -> Map {
    let width = cur[0].len();
    let height = cur.len();
    let mut cols: Board = vec![];
    for row in 0..width {
        cols.push(vec![]);
        for col in 0..height {
            cols[row].push(cur[col][row]);
        }
    }
    Map {
        rows: cur.clone(),
        cols,
        width,
        height,
    }
}

// don't forget to add an extra empty newline
fn parse(input: &str) -> Vec<Map> {
    let mut result: Vec<Map> = vec![];
    let mut cur: Board = vec![];
    for line in input.lines() {
        if line.is_empty() {
            if !cur.is_empty() {
                result.push(gen_map(&cur));
                cur.clear();
            }
            continue;
        }
        cur.push(
            line.chars()
                .filter_map(|t| (!t.is_ascii_whitespace()).then_some(t))
                .collect(),
        );
    }
    if !cur.is_empty() {
        result.push(gen_map(&cur));
    }
    result
}

pub fn part_one(input: &str) -> Option<u32> {
    Some(parse(input).iter().map(|t| t.part1()).sum())
}

pub fn part_two(input: &str) -> Option<u32> {
    Some(parse(input).iter().map(|t| t.part2()).sum())
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_part_one() {
        let result = part_one(&advent_of_code::template::read_file("examples", DAY));
        assert_eq!(result, Some(405));
    }

    #[test]
    fn test_part_two() {
        let result = part_two(&advent_of_code::template::read_file("examples", DAY));
        assert_eq!(result, Some(400));
    }

    #[test]
    fn sample_data2() {
        let result = part_one(&advent_of_code::template::read_file_part(
            "examples", DAY, 2,
        ));
        assert_eq!(result, Some(709));
        // assert_eq!(part2(&input), 1400);
    }
}
